__all__ = ("route", "HueAPIv2")

from abc import abstractmethod

from logging import getLogger
from re import compile as re_compile
from typing import Any, Generic, Iterable, Literal, Optional, TypeVar

from httpx import _content
from pydantic import BaseModel
from pydantic.generics import GenericModel
from ujson import dumps

from . import models
from .abc import SubRouter
from .models import UUID, HueEntsV2
from .utils import route

setattr(_content, "json_dumps", dumps)

UUID_CMP = re_compile(r"^[0-9a-f]{8}-(?:[0-9a-f]{4}-){3}[0-9a-f]{12}$")

logger = getLogger()

_T = TypeVar("_T")

TYPE_CACHE = {}

for k, v in HueEntsV2.__dict__.items():
    if k.startswith("__") or not issubclass(v, BaseModel):
        continue
    TYPE_CACHE[getattr(v, "type")] = v


class Event(GenericModel, Generic[_T]):
    class Config:
        __root__ = _T

    id: str
    object: _T
    type_: str


class HueEDK(SubRouter):
    BASE_URL = ""

    @route("GET", "/api/config")
    async def get_config(self, /):
        ...

    @route("POST", "/api")
    async def generate_token(
        self,
        /,
        **kwargs,
    ):
        ...

    @route("GET", "/api/v1")
    async def get_application_id(
        self,
        /,
        **kwargs,
    ):
        ...

    @route("GET", "/{script_path}")
    async def get_script_resource(self, script_path: str, /):
        ...


class HueAPIv1(SubRouter):
    BASE_URL = ""


class HueAPIv2(SubRouter):
    BASE_URI = "/clip/v2"

    @route("GET", "../../eventstream/clip/v2")
    async def listen_events(self, /):
        ...

    @route("GET", "/resource/behavior_instance")
    async def get_behavior_instances(self, /) -> list[HueEntsV2.BehaviorInstance]:
        ...

    @route("POST", "/resource/behavior_instance")
    async def create_behavior_instance(
        self,
        /,
        **kwargs,
    ) -> models.Attributes.Identifier:
        ...

    @route("GET", "/resource/behavior_instance/{behavior_instance_id}")
    async def get_behavior_instance(
        self,
        behavior_instance_id: UUID,
        /,
    ) -> HueEntsV2.BehaviorInstance:
        ...

    @route("PUT", "/resource/behavior_instance/{behavior_instance_id}")
    async def set_behavior_instance(
        self,
        behavior_instance_id: UUID,
        /,
        **kwargs,
    ) -> models.Attributes.Identifier:
        ...

    @route("DELETE", "/resource/behavior_instance/{behavior_instance_id}")
    async def delete_behavior_instance(
        self,
        behavior_instance_id: UUID,
        /,
    ) -> models.Attributes.Identifier:
        ...

    @route("GET", "/resource/behavior_script")
    async def get_behavior_scripts(self, /) -> list[HueEntsV2.BehaviorScript]:
        ...

    @route("GET", "/resource/behavior_script/{behavior_script_id}")
    async def get_behavior_script(
        self,
        behavior_script_id: UUID,
        /,
    ) -> HueEntsV2.BehaviorScript:
        ...

    @route("GET", "/resource/bridges")
    async def get_bridges(self, /) -> list[HueEntsV2.Bridge]:
        ...

    @route("GET", "/resource/bridges/{bridge_id}")
    async def get_bridge(
        self,
        bridge_id: UUID,
        /,
    ) -> HueEntsV2.Bridge:
        ...

    @route("PUT", "/resource/bridges/{bridge_id}")
    async def set_bridge(
        self,
        bridge_id: UUID,
        /,
        **kwargs,
    ) -> models.Attributes.Identifier:
        ...

    @route("GET", "/resource/bridge_home")
    async def get_bridge_homes(self, /) -> list[HueEntsV2.BridgeHome]:
        ...

    @route("GET", "/resource/bridge_home/{bridge_home_id}")
    async def get_bridge_home(
        self,
        bridge_home_id: UUID,
        /,
    ) -> HueEntsV2.BridgeHome:
        ...

    @route("PUT", "/resource/bridge_home/{bridge_home_id}")
    async def set_bridge_home(
        self,
        bridge_home_id: UUID,
        /,
        **kwargs,
    ) -> models.Attributes.Identifier:
        ...

    @route("GET", "/resource/button")
    async def get_buttons(self, /) -> list[HueEntsV2.Button]:
        ...

    @route("GET", "/resource/button/{button_id}")
    async def get_button(
        self,
        button_id: UUID,
        /,
    ) -> HueEntsV2.Button:
        ...

    @route("PUT", "/resource/button/{button_id}")
    async def set_button(
        self,
        button_id: UUID,
        /,
        **kwargs,
    ) -> models.Attributes.Identifier:
        ...

    @route("GET", "/resource/device")
    async def get_devices(self, /) -> list[Iterable[HueEntsV2.Device]]:
        ...

    @route("GET", "/resource/device/{device_id}")
    async def get_device(
        self,
        device_id: UUID,
        /,
    ) -> HueEntsV2.Device:
        ...

    @route("PUT", "/resource/device/{device_id}")
    async def set_device(
        self,
        device_id: UUID,
        /,
        **kwargs,
    ) -> models.Attributes.Identifier:
        ...

    @route("DELETE", "/resource/device/{device_id}")
    async def delete_device(
        self,
        device_id: UUID,
        /,
    ):
        ...

    @route("GET", "/resource/device_power")
    async def get_device_powers(self, /) -> list[HueEntsV2.DevicePower]:
        ...

    @route("GET", "/resource/device_power/{device_power_id}")
    async def get_device_power(
        self,
        device_power_id: UUID,
        /,
    ) -> HueEntsV2.DevicePower:
        ...

    @route("PUT", "/resource/device_power/{device_power_id}")
    async def set_device_power(
        self,
        device_power_id: UUID,
        /,
        **kwargs,
    ) -> models.Attributes.Identifier:
        ...

    @route("GET", "/resource/entertainment")
    async def get_entertainments(self, /) -> list[HueEntsV2.Entertainment]:
        ...

    @route("GET", "/resource/entertainment/{entertainment_id}")
    async def get_entertainment(
        self,
        entertainment_id: UUID,
        /,
    ) -> HueEntsV2.Entertainment:
        ...

    @route("PUT", "/resource/entertainment/{entertainment_id}")
    async def set_entertainment(
        self,
        entertainment_id: UUID,
        /,
        **kwargs,
    ) -> models.Attributes.Identifier:
        ...

    @route("GET", "/resource/entertainment_configuration")
    async def get_entertainment_configurations(
        self,
        /,
    ) -> list[HueEntsV2.EntertainmentConfiguration]:
        ...

    @route("POST", "/resource/entertainment_configuration")
    async def create_entertainment_configuration(
        self,
        /,
        **kwargs,
    ) -> models.Attributes.Identifier:
        ...

    @route(
        "GET", "/resource/entertainment_configuration/{entertainment_configuration_id}"
    )
    async def get_entertainment_configuration(
        self,
        entertainment_configuration_id: UUID,
        /,
    ) -> HueEntsV2.EntertainmentConfiguration:
        ...

    @route(
        "PUT", "/resource/entertainment_configuration/{entertainment_configuration_id}"
    )
    async def set_entertainment_configuration(
        self,
        entertainment_configuration_id: UUID,
        /,
        **kwargs,
    ) -> models.Attributes.Identifier:
        ...

    @route(
        "DELETE",
        "/resource/entertainment_configuration/{entertainment_configuration_id}",
    )
    async def delete_entertainment_configuration(
        self,
        entertainment_configuration_id: UUID,
        /,
    ) -> models.Attributes.Identifier:
        ...

    @route("GET", "/resource/geofence_client")
    async def get_geofence_clients(
        self,
        /,
    ) -> list[HueEntsV2.GeofenceClient]:
        ...

    @route("POST", "/resource/geofence_client")
    async def create_geofence_client(
        self,
        /,
        **kwargs,
    ) -> models.Attributes.Identifier:
        ...

    @route("GET", "/resource/geofence_client/{geofence_client_id}")
    async def get_geofence_client(
        self,
        geofence_client_id: UUID,
        /,
    ) -> HueEntsV2.GeofenceClient:
        ...

    @route("PUT", "/resource/geofence_client/{geofence_client_id}")
    async def set_geofence_client(
        self,
        geofence_client_id: UUID,
        /,
        **kwargs,
    ) -> models.Attributes.Identifier:
        ...

    @route("DELETE", "/resource/geofence_client/{geofence_client_id}")
    async def delete_geofence_client(
        self,
        geofence_client_id: UUID,
        /,
    ) -> models.Attributes.Identifier:
        ...

    @route("GET", "/resource/geolocation")
    async def get_geolocations(self, /) -> list[HueEntsV2.Geolocation]:
        ...

    @route("GET", "/resource/geolocation/{geolocation_id}")
    async def get_geolocation(
        self,
        geolocation_id: UUID,
        /,
    ) -> HueEntsV2.Geolocation:
        ...

    @route("PUT", "/resource/geolocation/{geolocation_id}")
    async def set_geolocation(
        self,
        geolocation_id: UUID,
        /,
        **kwargs,
    ) -> models.Attributes.Identifier:
        ...

    @route("GET", "/resource/grouped_light")
    async def get_grouped_lights(self, /) -> list[HueEntsV2.GroupedLight]:
        ...

    @route("GET", "/resource/grouped_light/{grouped_light_id}")
    async def get_grouped_light(
        self,
        grouped_light_id: UUID,
        /,
    ) -> HueEntsV2.GroupedLight:
        ...

    @route("PUT", "/resource/grouped_light/{grouped_light_id}")
    async def set_grouped_light(
        self,
        grouped_light_id: UUID,
        /,
        **kwargs,
    ) -> models.Attributes.Identifier:
        ...

    @route("GET", "/resource/homekit")
    async def get_homekits(self, /) -> list[HueEntsV2.Homekit]:
        ...

    @route("GET", "/resource/homekit/{homekit_id}")
    async def get_homekit(
        self,
        homekit_id: UUID,
        /,
    ) -> HueEntsV2.Homekit:
        ...

    @route("PUT", "/resource/homekit/{homekit_id}")
    async def set_homekit(
        self,
        homekit_id: UUID,
        /,
        type: Optional[str] = None,
        action: Optional[Literal["homekit_reset"]] = None,
    ) -> models.Attributes.Identifier:
        ...

    @route("GET", "/resource/light")
    async def get_lights(
        self,
        /,
        friendly_name: Optional[str] = None,
    ) -> list[HueEntsV2.Light]:
        ...

    @route("GET", "/resource/light/{light_id}")
    async def get_light(
        self,
        light_id: UUID,
        /,
    ) -> HueEntsV2.Light:
        ...

    @route("PUT", "/resource/light/{light_id}")
    async def set_light(
        self,
        light_id: UUID,
        /,
        on: Optional[dict[Literal["on"], bool]] = None,
        dimming: Optional[dict[str, Any]] = None,
        dimming_delta: Optional[dict] = None,
        color_temperature: Optional[int] = None,
        color_temperature_delta: Optional[dict] = None,
        color: Optional[dict] = None,
        dynamics: Optional[dict] = None,
        alert: Optional[dict] = None,
        gradient: Optional[dict] = None,
        effects: Optional[dict] = None,
        timed_effects: Optional[dict] = None,
    ) -> models.Attributes.Identifier:
        ...

    @route("GET", "/resource/light_level")
    async def get_light_levels(self, /) -> list[HueEntsV2.LightLevel]:
        ...

    @route("GET", "/resource/light_level/{light_level_id}")
    async def get_light_level(
        self,
        light_level_id: UUID,
        /,
    ) -> HueEntsV2.LightLevel:
        ...

    @route("PUT", "/resource/light_level/{light_level_id}")
    async def set_light_level(
        self,
        light_level_id: UUID,
        /,
        **kwargs,
    ) -> models.Attributes.Identifier:
        ...

    @route("GET", "/resource/matter")
    async def get_matters(self, /):
        ...

    @route("GET", "/resource/matter/{matter_id}")
    async def get_matter(
        self,
        matter_id: UUID,
        /,
    ):
        ...

    @route("PUT", "/resource/matter/{matter_id}")
    async def set_matter(
        self,
        matter_id: UUID,
        /,
        **kwargs,
    ):
        ...

    @route("GET", "/resource/matter_fabric")
    async def get_matter_fabrics(self, /):
        ...

    @route("GET", "/resource/matter_fabric/{matter_fabric_id}")
    async def get_matter_fabric(
        self,
        matter_fabric_id: UUID,
        /,
    ):
        ...

    @route("DELETE", "/resource/matter_fabric/{matter_fabric_id}")
    async def delete_matter_fabric(
        self,
        matter_fabric_id: UUID,
        /,
    ):
        ...

    @route("GET", "/resource/motion")
    async def get_motions(self, /) -> list[HueEntsV2.Motion]:
        ...

    @route("GET", "/resource/motion/{motion_id}")
    async def get_motion(
        self,
        motion_id: UUID,
        /,
    ) -> HueEntsV2.Motion:
        ...

    @route("PUT", "/resource/motion/{motion_id}")
    async def set_motion(
        self,
        motion_id: UUID,
        /,
        **kwargs,
    ) -> models.Attributes.Identifier:
        ...

    @route("GET", "/resource/relative_rotary")
    async def get_relative_rotaries(self, /) -> list[HueEntsV2.RelativeRotary]:
        ...

    @route("PUT", "/resource/relative_rotary/{relative_rotary_id}")
    async def set_relative_rotary(
        self,
        relative_rotary_id: UUID,
        /,
        **kwargs,
    ) -> models.Attributes.Identifier:
        ...

    @route("GET", "/resource/relative_rotary/{relative_rotary_id}")
    async def get_relative_rotary(
        self,
        relative_rotary_id: UUID,
        /,
    ) -> HueEntsV2.RelativeRotary:
        ...

    @route("POST", "/resource/relative_rotary")
    async def create_relative_rotary(
        self,
        /,
        **kwargs,
    ) -> models.Attributes.Identifier:
        ...

    @route("DELETE", "/resource/relative_rotary/{relative_rotary_id}")
    async def delete_relative_rotary(
        self,
        relative_rotary_id: UUID,
        /,
    ) -> models.Attributes.Identifier:
        ...

    @route("GET", "/resource/room")
    async def get_rooms(self, /) -> list[HueEntsV2.Room]:
        ...

    @route("POST", "/resource/room")
    async def create_room(
        self,
        /,
        **kwargs,
    ) -> models.Attributes.Identifier:
        ...

    @route("GET", "/resource/room/{room_id}")
    async def get_room(
        self,
        room_id: UUID,
        /,
    ) -> HueEntsV2.Room:
        ...

    @route("PUT", "/resource/room/{room_id}")
    async def set_room(
        self,
        room_id: UUID,
        /,
        metadata: Optional[dict[str, str]] = None,
        children: Optional[models.Attributes.Identifier] = None,
    ) -> models.Attributes.Identifier:
        ...

    @route("DELETE", "/resource/room/{room_id}")
    async def delete_room(
        self,
        room_id: UUID,
        /,
    ) -> models.Attributes.Identifier:
        ...

    @route("GET", "/resource/scene")
    async def get_scenes(self, /) -> list[HueEntsV2.Scene]:
        ...

    @route("POST", "/resource/scene")
    async def create_scene(
        self,
        /,
        **kwargs,
    ) -> models.Attributes.Identifier:
        ...

    @route("GET", "/resource/scene/{scene_id}")
    async def get_scene(
        self,
        scene_id: UUID,
        /,
    ) -> HueEntsV2.Scene:
        ...

    @route("PUT", "/resource/scene/{scene_id}")
    async def set_scene(
        self,
        scene_id: UUID,
        /,
        **kwargs,
    ) -> models.Attributes.Identifier:
        ...

    @route("DELETE", "/resource/scene/{scene_id}")
    async def delete_scene(
        self,
        scene_id: UUID,
        /,
    ) -> models.Attributes.Identifier:
        ...

    @route("GET", "/resource/smart_scene")
    async def get_smart_scenes(self, /):
        ...

    @route("POST", "/resource/smart_scene")
    async def create_smart_scene(
        self,
        /,
        **kwargs,
    ):
        ...

    @route("GET", "/resource/smart_scene/{smart_scene_id}")
    async def get_smart_scene(
        self,
        smart_scene_id: UUID,
        /,
    ):
        ...

    @route("PUT", "/resource/smart_scene/{smart_scene_id}")
    async def set_smart_scene(
        self,
        smart_scene_id: UUID,
        /,
        **kwargs,
    ):
        ...

    @route("DELETE", "/resource/smart_scene/{smart_scene_id}")
    async def delete_smart_scene(
        self,
        smart_scene_id: UUID,
        /,
    ):
        ...

    @route("GET", "/resource/temperature")
    async def get_temperatures(self, /) -> list[HueEntsV2.Temperature]:
        ...

    @route("GET", "/resource/temperature/{temperature_id}")
    async def get_temperature(
        self,
        temperature_id: UUID,
        /,
    ) -> HueEntsV2.Temperature:
        ...

    @route("PUT", "/resource/temperature/{temperature_id}")
    async def set_temperature(
        self,
        temperature_id: UUID,
        /,
        **kwargs,
    ) -> models.Attributes.Identifier:
        ...

    @route("GET", "/resource/zgb_connectivity")
    async def get_zgb_connectivities(self, /) -> list[HueEntsV2.ZGPConnectivity]:
        ...

    @route("GET", "/resource/zgb_connectivity/{zgb_connectivity_id}")
    async def get_zgb_connectivity(
        self,
        zgb_connectivity_id: UUID,
        /,
    ) -> HueEntsV2.ZGPConnectivity:
        ...

    @route("PUT", "/resource/zgb_connectivity/{zgb_connectivity_id}")
    async def set_zgb_connectivity(
        self,
        zgb_connectivity_id: UUID,
        /,
        **kwargs,
    ) -> models.Attributes.Identifier:
        ...

    @route("GET", "/resource/zigbee_connectivity")
    async def get_zigbee_connectivities(self, /) -> list[HueEntsV2.ZigbeeConnectivity]:
        ...

    @route("GET", "/resource/zigbee_connectivity/{zigbee_connectivity_id}")
    async def get_zigbee_connectivity(
        self,
        zigbee_connectivity_id: UUID,
        /,
    ) -> HueEntsV2.ZigbeeConnectivity:
        ...

    @route("PUT", "/resource/zigbee_connectivity/{zigbee_connectivity_id}")
    async def set_zigbee_connectivity(
        self,
        zigbee_connectivity_id: UUID,
        /,
        **kwargs,
    ) -> models.Attributes.Identifier:
        ...

    @route("GET", "/resource/zigbee_device_discovery")
    async def get_zigbee_device_discoveries(
        self,
        /,
    ) -> list[HueEntsV2.ZigbeeDeviceDiscovery]:
        ...

    @route("GET", " /resource/zigbee_device_discovery/{zigbee_device_discovery_id}")
    async def get_zigbee_device_discovery(
        self,
        zigbee_device_discovery_id: UUID,
        /,
    ) -> HueEntsV2.ZigbeeDeviceDiscovery:
        ...

    @route("PUT", "/resource/zigbee_device_discovery/{zigbee_device_discovery_id}")
    async def set_zigbee_device_discovery(
        self,
        zigbee_device_discovery_id: UUID,
        /,
        **kwargs,
    ) -> models.Attributes.Identifier:
        ...

    @route("GET", "/resource/zone")
    async def get_zones(self, /) -> list[HueEntsV2.Zone]:
        ...

    @route("POST", "/resource/zone")
    async def create_zone(
        self,
        /,
        **kwargs,
    ) -> models.Attributes.Identifier:
        ...

    @route("GET", "/resource/zone/{zone_id}")
    async def get_zone(
        self,
        zone_id: UUID,
        /,
    ) -> HueEntsV2.Zone:
        ...

    @route("PUT", "/resource/zone/{zone_id}")
    async def set_zone(
        self,
        zone_id: UUID,
        /,
        **kwargs,
    ) -> models.Attributes.Identifier:
        ...

    @route("DELETE", "/resource/zone/{zone_id}")
    async def delete_zone(
        self,
        zone_id: UUID,
        /,
    ) -> models.Attributes.Identifier:
        ...

    @route("GET", "/resource")
    async def get_resources(self, /) -> list[HueEntsV2.Resource]:
        ...

    @abstractmethod
    async def on_motion_update(
        self,
        motion: HueEntsV2.Motion,
    ):
        ...

    @abstractmethod
    async def on_button_update(
        self,
        button: HueEntsV2.Button,
    ):
        ...

    @abstractmethod
    async def on_zone_update(
        self,
        zone: HueEntsV2.Zone,
    ):
        ...

    @abstractmethod
    async def on_zigbee_connectivity_update(
        self,
        zigbee: HueEntsV2.ZigbeeConnectivity,
    ):
        ...

    @abstractmethod
    async def on_zgp_connectivity_update(
        self,
        zgp_connectivity: HueEntsV2.ZGPConnectivity,
    ):
        ...

    @abstractmethod
    async def on_temperature_update(
        self,
        temperature: HueEntsV2.Temperature,
    ):
        ...

    @abstractmethod
    async def on_scene_update(
        self,
        scene: HueEntsV2.Scene,
    ):
        ...

    @abstractmethod
    async def on_room_update(
        self,
        room: HueEntsV2.Room,
    ):
        ...

    @abstractmethod
    async def on_resource_update(
        self,
        device: HueEntsV2.Resource,
    ):
        ...

    @abstractmethod
    async def on_relative_rotary_update(
        self,
        relative_rotary: HueEntsV2.RelativeRotary,
    ):
        ...

    @abstractmethod
    async def on_light_level_update(
        self,
        light_level: HueEntsV2.LightLevel,
    ):
        ...

    @abstractmethod
    async def on_light_update(
        self,
        light: HueEntsV2.Light,
    ):
        ...

    @abstractmethod
    async def on_homekit_update(
        self,
        homekit: HueEntsV2.Homekit,
    ):
        ...

    @abstractmethod
    async def on_grouped_light_update(
        self,
        grouped_light: HueEntsV2.GroupedLight,
    ):
        ...

    @abstractmethod
    async def on_geolocation_update(
        self,
        geolocation: HueEntsV2.Geolocation,
    ):
        ...

    @abstractmethod
    async def on_geofence_client_update(
        self, geofence_client: HueEntsV2.GeofenceClient
    ):
        ...

    @abstractmethod
    async def on_entertainment_configuration_update(
        self,
        entertainment_configuration: HueEntsV2.EntertainmentConfiguration,
    ):
        ...

    @abstractmethod
    async def on_entertainment_update(
        self,
        entertainment: HueEntsV2.Entertainment,
    ):
        ...

    @abstractmethod
    async def on_device_power_update(
        self,
        device_power: HueEntsV2.DevicePower,
    ):
        ...

    @abstractmethod
    async def on_device_update(
        self,
        device: HueEntsV2.Device,
    ):
        ...

    @abstractmethod
    async def on_bridge_home_update(
        self,
        bridge_home: HueEntsV2.BridgeHome,
    ):
        ...

    @abstractmethod
    async def on_bridge_update(
        self,
        bridge: HueEntsV2.Bridge,
    ):
        ...

    @abstractmethod
    async def on_behavior_script_update(
        self,
        behavior_script: HueEntsV2.BehaviorScript,
    ):
        ...

    @abstractmethod
    async def on_behavior_instance_update(
        self,
        behavior_instance: HueEntsV2.BehaviorInstance,
    ):
        ...
